@@ -1,10 +1,10 @@
 /*
- * Raster3D V2.7
+ * Raster3D V3.0
  * local.c
  *
  * Output from render.f is performed by calls to routine LOCAL.
  *
- * This version of local.c supports 7 output modes, 5 of which
+ * This version of local.c supports 6 output modes, 5 of which
  * are controlled by conditional compilation directives.
  *
  *	mode 0	AVS image file sent to stdout
@@ -21,15 +21,11 @@
  *	mode 3	#ifdef TIFF_SUPPORT
  *		calls to the TIFF library if TIFF_SUPPORT is defined
  *
- *	mode 4	#ifdef JPEG_SUPPORT
+ *	mode 4	#ifdef GD_SUPPORT
  *		JPEG image output to file (defaults to stdout)
  *
- *	mode 5	#ifdef IMAGEPIPE
- *		AVS image file piped to ImageMagick for conversion to
- *		some other image type determined by suffix
- *
- *	mode 6	#ifdef PNG_SUPPORT
- *		PNG image output to file (defaults to stdout)
+ *	mode 5	#ifdef GD_SUPPORT
+ *		libgd PNG image output to file (defaults to stdout)
  *
  * Command line switches other than output file format handled elsewhere
  * here we recognize
@@ -39,12 +35,12 @@
  *		-avs  [filename] for AVS output
  *		-sgi  [filename] SGI libimage format output
  *		-tiff [filename] TIFF output format
- *		-out   file.xxx  pipe output to ImageMagick
  */
 
 #include	<stdio.h>
 #include	<fcntl.h>
 #include	<string.h>
+#include	<ctype.h>
 #include	<time.h>
 #include	<stdlib.h>
 #include	<netinet/in.h>
@@ -56,33 +52,13 @@
 #ifdef TIFF_SUPPORT
 #include        <tiff.h>
 #include        <tiffio.h>
-#ifdef LINUX
+#if defined(LINUX) || defined(OSX)
 #include	<signal.h>
-#endif /* LINUX */
+#endif
 #endif /* TIFF_SUPPORT */
 
-#ifdef JPEG_SUPPORT
-#include	<jpeglib.h>
-#include	<jerror.h>
-#ifdef BROKEN_LIBJPEG
-#undef jpeg_create_compress
-#endif
-  /* JPEG data structures */
-  static struct jpeg_compress_struct	cinfo;
-  static struct jpeg_error_mgr      	jerr;
-  static JSAMPLE			*jpeg_pixels;
-  static JSAMPROW			jpeg_row[1];
-#endif
-
-#ifdef PNG_SUPPORT
-#include	<png.h>
-#ifndef png_jmpbuf
-#  define png_jmpbuf(png_ptr) ((png_ptr)->jmpbuf)
-#endif
-  /* PNG data structures */
-  static png_structp png_ptr;
-  static png_infop   info_ptr;
-  static png_text    text_ptr[2];
+#if defined GD_SUPPORT
+#include "gd.h"
 #endif
 
 /* Pick up version number from same include file used by Makefile */
@@ -100,6 +76,18 @@ char *
 
 int		alpha_channel = 0;
 
+/* Loaded by mode 5; used to initialize image file by mode 1 */
+static int bkg_r = 0, bkg_g = 0, bkg_b = 0;
+
+/* Define bits to be pass to both the local_() and addlabel_() functions */
+static int xsize = 0, ysize = 0;
+static int mode = -1; /* this is needed by goth local_() and addlabel_() */
+#if defined GD_SUPPORT
+  /* For -png OR -jpeg (using libgd) output option only */
+  static gdImagePtr molecule_img = NULL;
+  static gdImagePtr label_img = NULL;
+#endif
+
 /* HPUX lacks Fortran intrinsic functions AND and OR for some reason, */
 /* so I put a copy here. On the other hand HPUX has an unusually sane */
 /* calling convention for Fortran subroutine names.                   */
@@ -114,6 +102,7 @@ int and_(i,j) int *i,*j; {return (*i & *j);}
 int or_(i,j)  int *i,*j; {return (*i | *j);}
 #endif
 
+size_t trimwhitespace(char *out, size_t len, const char *str);
 
 int local_(option,buffer1,buffer2,buffer3,buffer4)
      int	*option;
@@ -121,14 +110,12 @@ int local_(option,buffer1,buffer2,buffer3,buffer4)
 {
   
   /* Everyone needs these */
-  static int	xsize, ysize;
-  static int	mode    = -1;
   int	        i;
   static char	*ofile;
   int		status = 0;
   int		invert = 0;
   int		bits;
-  int		quality;
+  static int	quality = 90;
  
   static time_t	start_time, end_time;
   static char program_name[20] = "Raster3D         G";
@@ -151,25 +138,6 @@ int local_(option,buffer1,buffer2,buffer3,buffer4)
   void my_write_tiff();
 #endif
 
-#ifdef JPEG_SUPPORT
-  /* For -jpeg output option only */
-  JSAMPLE *q;
-  static FILE *jpegfile;
-#endif
-
-#ifdef IMAGEPIPE
-  /* For -out output option only */
-  static FILE *pipe;
-  char convert_command[128];
-#endif
-
-#ifdef PNG_SUPPORT
-  /* For -png output option only */
-  static png_bytep pngline;
-  static FILE *pngfile;
-#endif
-
-
 
 /****************************************************************/
 /* The action taken by this subroutine is determined by the	*/
@@ -203,14 +171,13 @@ if (*option == 0)
       }
 
     /* Version 2.7a - AVS used to be the default, but no longer */
-    else if (strncmp( (char *)buffer1, "-avs" , 4) == 0)
-      {
-	mode  = 0;
-      }
+    else if (strncmp( (char *)buffer1, "-avs" , 4) == 0) {
+	mode = 0;
+    }
 
     else if (strncmp( (char *)buffer1, "-sgi" , 4) == 0)
       {
-#ifdef      LIBIMAGE_SUPPORT
+#ifdef LIBIMAGE_SUPPORT
 	mode  = 2;
 	ofile = (char *)buffer2;
 	invert = !invert;
@@ -223,49 +190,32 @@ if (*option == 0)
 
     else if (strncmp( (char *)buffer1, "-jpeg" , 5) == 0)
       {
-#ifdef      JPEG_SUPPORT
+#ifdef GD_SUPPORT /* JPEG */
 	mode  = 4;
 	ofile = (char *)buffer2;
 #else
 	fprintf(stderr,
-		"\n This copy of render was not built with jpeg support\n");
+		"\n This copy of render was not built with libgd jpeg support\n");
 	exit(-1);
 #endif
       }
 
     else if (strncmp( (char *)buffer1, "-png" , 4) == 0)
       {
-#ifdef      PNG_SUPPORT
-	mode  = 6;
+#ifdef GD_SUPPORT /* PNG */
+	mode  = 5;
 	ofile = (char *)buffer2;
 #else
 	fprintf(stderr,
-		"\n This copy of render was not built with png support\n");
-	exit(-1);
-#endif
-      }
-
-    else if (strncmp( (char *)buffer1, "-out", 4) == 0)
-      {
-#ifdef	IMAGEPIPE
-	mode  = 5;
-	ofile = strtok( (char *)buffer2, " " );
-	if (ofile)
-		fprintf(stderr,"ofile = %s\n",ofile);
-	else
-		fprintf(stderr,"ofile = NULL\n");
-#else
-	fprintf(stderr,
-		"\n This copy of render was not built with support\n");
-	fprintf(stderr,
-		" for piping output to ImageMagick (-out file)\n");
+		"\n This copy of render was not built with libgd png support\n");
 	exit(-1);
 #endif
       }
 
+    /* This option is long since deprecated */
     else if (strncmp( (char *)buffer1, "-orig", 5) == 0)
       {
-      	mode  = 1;
+      	mode = 1;
 	invert = !invert;
       }
     else if (strncmp( (char *)buffer1, "  ", 2) != 0)
@@ -276,17 +226,19 @@ if (*option == 0)
 	fprintf(stderr, "\n\n Usage:");
 	fprintf(stderr, "\n   input from stdin; output mode controlled from command line \n");
 	fprintf(stderr,
-		"\n     render [-png [outfile]]       PNG image to stdout (default!) or file");
+		"\n     render [-png [outfile]]       PNG image to stdout (default) or file");
+	fprintf(stderr,
+		"\n     render -jpeg [outfile]        JPEG image to stdour (default) or file");
 	fprintf(stderr,
 		"\n     render -avs                   AVS image to stdout");
-	fprintf(stderr,
-		"\n     render -jpeg [outfile]        JPEG image to outfile (defaults to stdout)");
+#ifdef LIBIMAGE_SUPPORT
 	fprintf(stderr,
 		"\n     render -sgi  [outfile]        output to SGI libimage file (defaults to render.rgb)");
+#endif
+#ifdef TIFF_SUPPORT
 	fprintf(stderr,
 		"\n     render -tiff [outfile]        output to TIFF file (defaults to render.tif)");
-	fprintf(stderr,
-		"\n     render -out outfile.xxx       pipe output via ImageMagick to get image type xxx");
+#endif
 	fprintf(stderr,"\n");
 	fprintf(stderr,"\n Options:");
 	fprintf(stderr,"\n   these over-ride contents of input stream header records \n");
@@ -295,10 +247,10 @@ if (*option == 0)
 	fprintf(stderr,"\n    -bg white|black|<col> set background color (<col> is hex #RRGGBB)");
 	fprintf(stderr,"\n    -debug                verbose output while running");
 	fprintf(stderr,"\n    -draft                no anti-aliasing (SCHEME 1)");
-	fprintf(stderr,"\n    -fontscale FF         multiplier for PostScript font size");
+	fprintf(stderr,"\n    -fontscale FF         multiplier for libgd font size [default 1.0]");
 	fprintf(stderr,"\n    -gamma GG             gamma correction applied to output image");
 	fprintf(stderr,"\n    -invert               invert y axis");
-	fprintf(stderr,"\n    -labels               write labels to PostScript file label3d.ps");
+	fprintf(stderr,"\n    -labels               pass labels to libgd to composite with molecular image");
 	fprintf(stderr,"\n    -quality QQ           0 < QQ < 95  jpeg compression [default 90]");
 	fprintf(stderr,"\n    -[no]shadow           enable or disable shadowing");
 	fprintf(stderr,"\n    -size HHHxVVV         specify size of output image in pixels");
@@ -308,14 +260,14 @@ if (*option == 0)
 	exit(-1);
       }
 
-    else /* Version 2.7a - default to png */
+    else /* default to png */
       {
-#ifdef      PNG_SUPPORT
-	mode  = 6;
+#ifdef GD_SUPPORT /* PNG */
+	mode  = 5;
 	ofile = (char *)buffer2;
 #else
 	fprintf(stderr,
-		"\n This copy of render was not built with png support\n");
+		"\n This copy of render was not built with libgd png support\n");
 	fprintf(stderr,
 		"Defaulting to AVS instead\n");
 	mode  = 0;
@@ -342,10 +294,10 @@ if (mode < 0)
 /****************************************************************/
 else if (*option == 1)
     {
-    xsize   = *buffer1;
-    ysize   = *buffer2;
-    bits    = *buffer3;
-    quality = *buffer4;
+    xsize   = *(int *)buffer1;
+    ysize   = *(int *)buffer2;
+    bits    = *(int *)buffer3;
+    quality = *(int *)buffer4;
 
     if (bits & ALPHACHANNEL) alpha_channel = 1;
       
@@ -369,7 +321,7 @@ else if (*option == 1)
 	}
     else
 
-#ifdef      LIBIMAGE_SUPPORT
+#ifdef LIBIMAGE_SUPPORT
     if (mode == 2)   /* sgi rgb mode */
 	{
 	if (*ofile != ' ')
@@ -429,116 +381,30 @@ else if (*option == 1)
     else
 #endif
 
-#ifdef	JPEG_SUPPORT
-    if (mode == 4)
+#ifdef GD_SUPPORT /* JPEG + PNG */
+    if (mode == 4 || mode == 5) {
+	int transparent;
+	molecule_img = gdImageCreateTrueColor(xsize, ysize);
+	label_img = gdImageCreateTrueColor(xsize, ysize);
+	gdImageAlphaBlending(label_img, 0);
+/*
+	if (!molecule_img)
 	{
-	if (*ofile != ' ' && *ofile != '-')
-	    {
-	    ofile = strtok( ofile, " " );
-	    jpegfile = fopen(ofile, "wb");
-	    if (!jpegfile)
-	    	{
-		fprintf(stderr,"Could not open output file %s\n",ofile);
-		exit(-1);
-		}
-	    }
-	else
-	    jpegfile = stdout;
-	jpeg_pixels = (JSAMPLE *) malloc(xsize*3*sizeof(JSAMPLE));
-	if (jpeg_pixels == (JSAMPLE *) NULL)
-	    {
-	      fprintf(stderr,"\nMemory allocation error\n");
-	      return(-1);
-	    }
-	cinfo.err = jpeg_std_error(&jerr);
-	jpeg_create_compress(&cinfo);
-	jpeg_stdio_dest(&cinfo,jpegfile);
-	cinfo.image_width      = xsize;
-	cinfo.image_height     = ysize;
-	cinfo.input_components = 3;
-	cinfo.in_color_space   = JCS_RGB;
-	jpeg_set_defaults(&cinfo);
-	jpeg_set_quality(&cinfo, quality, TRUE );
-
-	/* The following choices affect efficiency on a given machine */
-/*	cinfo.optimize_coding	= TRUE; */
-	cinfo.dct_method	= JDCT_FLOAT;
-
-	jpeg_simple_progression(&cinfo);
-	jpeg_start_compress(&cinfo, TRUE);
-	}
-    else
-#endif
-	
-#ifdef	IMAGEPIPE
-    if (mode == 5)   /* pipe to ImageMagick */
-	{
-	if (!ofile || *ofile==' ' || *ofile=='-') 
-	    { 
-	    fprintf(stderr,"render: illegal output file name\n");
-	    exit(-1);
-	    }
-	sprintf( convert_command, "convert -verbose -quality %d avs:- ", 
-		 quality );
-	strncat( convert_command, ofile, 96 );
-	fprintf(stderr,"\n Opening pipe to | %s\n",convert_command);
-	pipe = popen( convert_command, "w" );
-	if (!pipe) 
-	    {
-	    fprintf(stderr," pipe failed - die\n");
-	    exit(-1);
-	    }
-#ifdef NETWORKBYTEORDER
-	putw( htonl(xsize), pipe );
-	putw( htonl(ysize), pipe );
-#else
-	putw( xsize, pipe );
-	putw( ysize, pipe );
-#endif
-	fflush(pipe);
-	}
-    else
-#endif
-
-#ifdef PNG_SUPPORT
-    if (mode == 6) 
-    	{
-	if (*ofile != ' ' && *ofile != '-')
-	    {
-	    ofile = strtok( ofile, " " );
-	    pngfile = fopen(ofile, "wb");
-	    if (!pngfile)
-	    	{
-		fprintf(stderr,"Could not open output file %s\n",ofile);
-		exit(-1);
-		}
-	    }
-	else
-	    pngfile = stdout;
-	png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
-	if (png_ptr) info_ptr = png_create_info_struct(png_ptr);
-	if (!png_ptr || !info_ptr)
-	    {
 	    fprintf(stderr,"PNG initialization failed - die\n");
 	    exit(-1);
-	    }
-	if (setjmp(png_jmpbuf(png_ptr)))
-	    {
-	    fprintf(stderr,"Error writing PNG file - die\n");
-	    exit(-1);
-	    }
-	png_init_io(png_ptr,pngfile);
-	png_set_IHDR(png_ptr, info_ptr, (png_uint_32)xsize, (png_uint_32)ysize, 8, 
-		     (alpha_channel ? PNG_COLOR_TYPE_RGB_ALPHA : PNG_COLOR_TYPE_RGB),
-		     PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
-	pngline=(png_bytep) calloc(xsize, sizeof(png_byte) * (alpha_channel ? 4 : 3));
-	if (!pngline)
-	    {
-	    fprintf(stderr,"\nMemory allocation error\n");
-	    return(-1);
-	    }
-	if (quality == 100) png_set_compression_level( png_ptr, Z_NO_COMPRESSION );
 	}
+*/
+
+        /* Signal that all freetype font calls in this program will receive
+         * fontconfig patterns rather than filenames of font files */
+        gdFTUseFontConfig(1);
+
+	transparent = gdImageColorResolve(label_img, bkg_r, bkg_g, bkg_b);
+        gdImageColorTransparent(label_img, transparent);
+	gdImageFilledRectangle(label_img,0,0,xsize,ysize,transparent);
+
+	ofile = strtok( ofile, " " );
+    }
     else
 #endif
 
@@ -597,74 +463,53 @@ else if (*option == 2)
 	}
     else
 #endif
-	  
+
 #ifdef TIFF_SUPPORT
-    if (mode ==3)
+    if (mode == 3)
 	{
 	my_write_tiff(tfile,buffer1,buffer2,buffer3,buffer4,xsize,scanline);
 	}
     else
 #endif
 
-#ifdef JPEG_SUPPORT
-    if (mode == 4)
-	{
-	jpeg_row[0] = (JSAMPROW) jpeg_pixels;
-	q = jpeg_pixels;
-	for (i=0; i<xsize; i++) 
-	    {
-	    *q++ = (JSAMPLE) buffer1[i];
-	    *q++ = (JSAMPLE) buffer2[i];
-	    *q++ = (JSAMPLE) buffer3[i];
-	    }
-	jpeg_write_scanlines( &cinfo, jpeg_row, 1 );
-	}
-    else
-#endif 
-	
-#ifdef IMAGEPIPE
-    if (mode == 5) 	/* out pipe to ImageMagick */
+#ifdef GD_SUPPORT /* JPEG + PNG */
+    if (mode == 4 || mode == 5)
 	{
+/* TODO: Speed up using full-line-at-a-time copy instead of pixel-by-pixel */
+	int pixel_color;
+	static int j = 0;
+	int transparent;
+
 	if (alpha_channel)
-	    for (i=0; i<xsize; i++)
-		{
-		putc( buffer4[i], pipe );
-		putc( buffer1[i], pipe );
-		putc( buffer2[i], pipe );
-		putc( buffer3[i], pipe );
-		}
-	else
-	    for (i=0; i<xsize; i++)
-		{
-		putc( 255, pipe );
-		putc( buffer1[i], pipe );
-		putc( buffer2[i], pipe );
-		putc( buffer3[i], pipe );
-		}
-	}
-    else
+#if 1
+	    for (i=0; i<xsize; i++) {
+		pixel_color = gdImageColorResolveAlpha(molecule_img,
+			buffer1[i], buffer2[i], buffer3[i], 
+			(255-buffer4[i])>>1); /* inverse bitwise shift */
+		gdImageSetPixel(molecule_img, i, j, pixel_color);
+	    }
+
+/* TODO: Check the following to see if it is faster than the above. */
+#else
+	    for (i=0; i<xsize; i++) {
+		molecule_img->tpixels[j][i] = gdTrueColorAlpha(
+			buffer1[i], buffer2[i], buffer3[i], 
+			(255-buffer4[i])>>1);
+	    }
 #endif
-  
-#ifdef PNG_SUPPORT
-    if (mode == 6)
-	{
-	int j = 0;
-
-	if (alpha_channel)
-	    for (i=0; i<xsize; i++) {
-		pngline[j++] = buffer1[i];
-		pngline[j++] = buffer2[i];
-		pngline[j++] = buffer3[i];
-		pngline[j++] = buffer4[i];
-	    }
 	else
 	    for (i=0; i<xsize; i++) {
-		pngline[j++] = buffer1[i];
-		pngline[j++] = buffer2[i];
-		pngline[j++] = buffer3[i];
+		pixel_color = gdImageColorResolve(molecule_img,
+			buffer1[i], buffer2[i], buffer3[i]);
+		gdImageSetPixel(molecule_img, i, j, pixel_color);
 	    }
-	png_write_rows(png_ptr, &pngline, 1);
 
+	if (alpha_channel) {
+	    transparent = gdImageColorResolveAlpha(molecule_img, 0, 0, 0, 127);
+	    gdImageColorTransparent(molecule_img, transparent);
+	}
+
+	j++;
 	}
     else
 #endif
@@ -694,38 +539,50 @@ else if (*option == 3)
     if (mode == 3)
 	{
 	(void) TIFFFlushData(tfile);
-#ifdef LINUX
+#if defined(LINUX) || defined(OSX)
 	signal( SIGSEGV, SIG_IGN );
 #endif /* LINUX */
 	(void) TIFFClose(tfile);
 	}
     else
 #endif
-      
-#ifdef JPEG_SUPPORT
-    if (mode == 4)
-	{
-	jpeg_finish_compress(&cinfo);
-	}
-    else
-#endif
-  
-#ifdef IMAGEPIPE
-    if (mode == 5)
-	{
-	fflush(pipe);
-	pclose(pipe);
-        }
-    else
-#endif
 
-#ifdef PNG_SUPPORT
-    if (mode == 6)
-    	{
-	png_write_end(png_ptr, info_ptr);
-	png_destroy_write_struct(&png_ptr, &info_ptr);
-	free(pngline);
+#ifdef GD_SUPPORT /* JPEG + PNG */
+    if (mode == 4 || mode == 5) {
+
+	if (label_img) {
+	    /*
+	    gdImageAlphaBlending(molecule_img, 0);
+	    gdImageSaveAlpha(molecule_img, 1);
+	    */
+	    gdImageSetTile(molecule_img, label_img);
+	    gdImageFilledRectangle(molecule_img, 0, 0, xsize, ysize, gdTiled);
 	}
+
+	/*if (*ofile != ' ' && *ofile != '-')*/
+	if (ofile) {
+	    FILE *fd = fopen(ofile,"wb");
+            if (!fd) {
+                fprintf(stderr,"Could not open output file %s\n",ofile);
+                exit(-1);
+            }
+	    if (mode == 4)
+		gdImageJpeg(molecule_img, fd, quality);
+	    else
+		gdImagePng(molecule_img, fd);
+	    fclose(fd);
+
+	    if (label_img) {
+		gdImageDestroy(label_img);
+	    }
+	}
+	else
+	    if (mode == 4)
+		gdImageJpeg(molecule_img, stdout, quality);
+	    else
+		gdImagePng(molecule_img, stdout);
+	gdImageDestroy(molecule_img);
+    }
     else
 #endif
 
@@ -754,37 +611,26 @@ else if (*option == 4)
     else
 #endif
     
-#ifdef JPEG_SUPPORT
-    if (mode == 4)
-	{
-	jpeg_write_marker(&cinfo,JPEG_COM,(unsigned char *)program_name,strlen(program_name));
-	jpeg_write_marker(&cinfo,JPEG_COM,(unsigned char *)"\n",2);
-	jpeg_write_marker(&cinfo,JPEG_COM,(unsigned char *)buffer1,80);
-	}
-    else
-#endif
-
-#ifdef PNG_SUPPORT
-    if (mode == 6)
-    	{
-	text_ptr[0].key  = "Software";
-	text_ptr[0].text = (unsigned char *)program_name;
-	text_ptr[0].compression = PNG_TEXT_COMPRESSION_NONE;
-	text_ptr[0].text_length = strlen(program_name);
-	text_ptr[1].key  = "Title";
-	text_ptr[1].text = (unsigned char *)buffer1;
-	text_ptr[1].compression = PNG_TEXT_COMPRESSION_NONE;
-	text_ptr[1].text_length = 79;
-	png_set_text(png_ptr, info_ptr, text_ptr, 2);
-	png_write_info(png_ptr, info_ptr);
-	}
-    else
-#endif
-
     /* NOP */ ;
     return(1);
     }
-  
+
+/****************************************************************/
+/* Load background color to local storage			*/
+/****************************************************************/
+else if (*option == 5)
+    {
+    
+    bkg_r  = *buffer1;
+    bkg_g  = *buffer2;
+    bkg_b  = *buffer3;
+/*
+    fprintf(stderr,"Loading background color %d %d %d\n", bkg_r, bkg_g, bkg_b);
+ */
+    return(1);
+    }
+
+return 0;
 }
 
 #ifdef TIFF_SUPPORT
@@ -818,3 +664,165 @@ int j = 0;
 row++;
 }
 #endif
+
+/*===========================================================================*/
+/*
+ * _addlabel()
+ */
+int addlabel_(fontname, fontsize, fontscale, fontalign, xp, yp, zp, red, grn, blu,
+	labelstring, font_len, label_len)
+	int *fontalign;
+	float *fontsize, *fontscale, *xp, *yp, *zp, *red, *grn, *blu;
+	char *fontname, *labelstring;
+	long int font_len, label_len;
+{
+#ifdef GD_SUPPORT
+	static double last_x = 0.0, last_y = 0.0, last_z = 0.0;
+
+	double x = *xp, y = *yp, z = *zp;
+        float angle = 0.0;
+	int brect[8], font_color;
+	char *err, font[128], instring[128], *justify;
+	size_t out_size; /* out_size is just a dummy variable */
+	char *string;
+
+	if (mode != 4 && mode != 5)
+	    return 0;
+
+	/* TODO: If FONTNAME does not exist in the GDFONTPATH, fail gracefully */
+	out_size = trimwhitespace(font, 128, fontname);
+	out_size = trimwhitespace(instring, 128, labelstring);
+	string = instring;
+
+	/* Allocate colours */
+	/* FIXME: how to pass non-zero alpha value? */
+	font_color = gdImageColorAllocateAlpha(label_img, 
+	    (double)*red*255.0, (double)*grn*255.0, (double)*blu*255.0, 0);
+
+        /* Draw once with a NULL image to get the bounding rectangle */
+        err = gdImageStringFT(NULL, &brect[0], font_color,
+		font, (double)*fontsize*(double)*fontscale, angle, 0, 0, string);
+	if (err)
+		fprintf(stderr,"gdImageStringFT: %s while printing string %s with font %s\n",
+			err, string, font);
+
+	/* Align the string */
+	if ((int)*fontalign == 1) {
+		/* Justify string center */
+		justify = "Center";
+		x -= (brect[2]-brect[0]) / 2.;
+		y += (brect[3]-brect[1]) / 2.;
+	}
+	else if ((int)*fontalign == 2) {
+		/* Justify string right */
+		justify = "Right";
+		x -= (brect[2]-brect[0]);
+		y += (brect[3]-brect[1]);
+	}
+	else if ((int)*fontalign == 3) {
+		justify = "Offset";
+		x += last_x;
+		y += last_y;
+	}
+	else { 
+		/* Do nothing (default) */ 
+		justify = "Left";
+	}
+
+	/* TODO: Process super/subscripts */
+	while (strpbrk(string, "{}^_")) {
+	    char *mark = strpbrk(string, "{}^_");
+	    char save_mark = *mark;
+
+	    /* dump up to the first special character */
+	    *mark = '\0';
+	    err = gdImageStringFT(label_img, &brect[0], font_color,
+		font, (double)*fontsize*(double)*fontscale, angle, x, y, string);
+	    x += (brect[2] - brect[0]);
+	    string = mark+1;
+
+	    /* start sub/superscript placement */
+	    last_y = y;
+	    if (save_mark == '^')
+		y -= 0.5 * (*fontsize) * (*fontscale);
+	    else if (save_mark == '_')
+		y += 0.5 * (*fontsize) * (*fontscale);
+	    /* dump either a single character (careful: UTF-8!) */
+	    /* or the string enclosed by {} */
+	    if (*string == '{') {
+		string++;
+		if ((mark = strchr(string,'}')) != NULL) {
+		    *mark = '\0';
+		    err = gdImageStringFT(label_img, &brect[0], font_color,
+			font, (double)*fontsize*(double)*fontscale, angle, x, y, string);
+		    x += (brect[2] - brect[0]);
+		    y = last_y;
+		    string = mark+1;
+		} else {
+		    /* No closing '}', just print the rest of the string */
+		}
+
+	    } else {
+	    	/* FIXME: check if it's not really UTF-8 */
+		char utf8[6] = {0,0,0,0,0,0};
+		char *u = utf8;
+		*u = *string++;
+		if ((*u & 0xC0) == 0xC0) { /* Start of a UTF-8 sequence */
+		    *++u = *string++;
+		    if ((*string & 0xc0) == 0x80)
+			*++u = *string++;
+		    if ((*string & 0xc0) == 0x80)
+			*++u = *string++;
+		    if ((*string & 0xc0) == 0x80)
+			*++u = *string++;
+		}
+		err = gdImageStringFT(label_img, &brect[0], font_color,
+			font, (double)*fontsize*(double)*fontscale, angle, x, y, utf8);
+		x += (brect[2] - brect[0]);
+		y = last_y;
+	    }
+
+	}
+
+	/* Now render the remaining string, if any */
+	err = gdImageStringFT(label_img, &brect[0], font_color,
+		font, (double)*fontsize*(double)*fontscale, angle, x, y, string);
+	if (err)
+		fprintf(stderr,"gdImageStringFT: %s while printing string %s with font %s\n",
+			err, string, font);
+
+	last_x = x + (brect[2] - brect[0]);
+	last_y = y;
+	last_z = z;
+
+#endif
+	return 0;
+}
+
+size_t trimwhitespace(char *out, size_t len, const char *str)
+{
+	/* A simple function to trim the whitespace from the right-side of
+	 * a string.
+	 */
+
+	const char *end;
+	size_t out_size;
+
+	/* Trim leading space */
+	/*while(isspace(*str)) str++;*/
+
+	/* Trim trailing space */
+	end = str + strlen(str) - 1;
+	while(end > str && isspace(*end)) end--;
+	end++;
+
+	/* Set output size to minimum of trimmed string length and buffer size
+	 * minus 1 */
+	out_size = (end - str) < len-1 ? (end - str) : len-1;
+
+	/* Copy trimmed string and add null terminator */
+	memcpy(out, str, out_size);
+	out[out_size] = 0;
+
+	return out_size;
+}
